@@ -1,6 +1,7 @@
 /*
- *  Copyright (C) 2017 Savoir-faire Linux Inc.
- *  Author : Sébastien Blin <sebastien.blin@savoirfairelinux.com>
+ *  Copyright (C) 2017-2018 Savoir-faire Linux Inc.
+ *  Author: Sébastien Blin <sebastien.blin@savoirfairelinux.com>
+ *          Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,11 +24,19 @@
 #include "def.h"
 #include "sockaddr.h"
 #include "infohash.h"
+#include "scheduler.h"
+#include "value.h"
+#include "proxy.h"
 
 #include <thread>
 #include <memory>
+#include <mutex>
 #include <restbed>
 
+namespace Json {
+    class Value;
+}
+
 namespace dht {
 
 class DhtRunner;
@@ -42,10 +51,11 @@ public:
      * Start the Http server for OpenDHT
      * @param dht the DhtRunner linked to this proxy server
      * @param port to listen
+     * @param pushServer where to push notifications
      * @note if the server fails to start (if port is already used or reserved),
      * it will fails silently
      */
-    DhtProxyServer(std::shared_ptr<DhtRunner> dht, in_port_t port = 8000);
+    DhtProxyServer(std::shared_ptr<DhtRunner> dht, in_port_t port = 8000, const std::string& pushServer = "");
     virtual ~DhtProxyServer();
 
     DhtProxyServer(const DhtProxyServer& other) = delete;
@@ -53,6 +63,27 @@ public:
     DhtProxyServer& operator=(const DhtProxyServer& other) = delete;
     DhtProxyServer& operator=(DhtProxyServer&& other) = delete;
 
+    struct ServerStats {
+        /** Current number of listen operations */
+        size_t listenCount;
+        /** Current number of permanent put operations */
+        size_t putCount;
+        /** Current number of push tokens with at least one listen operation */
+        size_t pushListenersCount;
+        /** Average requests per second */
+        double requestRate;
+
+        std::string toString() const {
+            std::ostringstream ss;
+            ss << "Listens: " << listenCount << " Puts: " << putCount << " PushListeners: " << pushListenersCount << std::endl;
+            ss << "Requests: " << requestRate << " per second." << std::endl;
+            return ss.str();
+        }
+    };
+    ServerStats getStats() const;
+
+    std::shared_ptr<DhtRunner> getNode() const { return dht_; }
+
     /**
      * Stop the DhtProxyServer
      */
@@ -69,7 +100,7 @@ private:
     void getNodeInfo(const std::shared_ptr<restbed::Session>& session) const;
 
     /**
-     * Return Values of a InfoHash
+     * Return Values of an infoHash
      * Method: GET "/{InfoHash: .*}"
      * Return: Multiple JSON object in parts. Example:
      * Value in JSON format\n
@@ -81,7 +112,7 @@ private:
     void get(const std::shared_ptr<restbed::Session>& session) const;
 
     /**
-     * Listen incoming Values of a InfoHash.
+     * Listen incoming Values of an infoHash.
      * Method: LISTEN "/{InfoHash: .*}"
      * Return: Multiple JSON object in parts. Example:
      * Value in JSON format\n
@@ -90,7 +121,7 @@ private:
      * On error: HTTP 503, body: {"err":"xxxx"}
      * @param session
      */
-    void listen(const std::shared_ptr<restbed::Session>& session) const;
+    void listen(const std::shared_ptr<restbed::Session>& session);
 
     /**
      * Put a value on the DHT
@@ -101,7 +132,9 @@ private:
      * HTTP 400, body: {"err":"xxxx"} if bad json or HTTP 502 if put fails
      * @param session
      */
-    void put(const std::shared_ptr<restbed::Session>& session) const;
+    void put(const std::shared_ptr<restbed::Session>& session);
+
+    void cancelPut(const InfoHash& key, Value::Id vid);
 
 #if OPENDHT_PROXY_SERVER_IDENTITY
     /**
@@ -128,7 +161,7 @@ private:
 #endif // OPENDHT_PROXY_SERVER_IDENTITY
 
     /**
-     * Return Values of a InfoHash filtered by a value id
+     * Return Values of an infoHash filtered by a value id
      * Method: GET "/{InfoHash: .*}/{ValueId: .*}"
      * Return: Multiple JSON object in parts. Example:
      * Value in JSON format\n
@@ -144,13 +177,61 @@ private:
      * Method: OPTIONS "/{hash: .*}"
      * Return: HTTP 200 + Allow: allowed methods
      * See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS
+     * @param session
      */
     void handleOptionsMethod(const std::shared_ptr<restbed::Session>& session) const;
 
+    /**
+     * Remove finished listeners
+     * @param testSession if we remove the listener only if the session is closed
+     */
+    void removeClosedListeners(bool testSession = true);
+
+#if OPENDHT_PUSH_NOTIFICATIONS
+    /**
+     * Subscribe to push notifications for an iOS or Android device.
+     * Method: SUBSCRIBE "/{InfoHash: .*}"
+     * Body: {"key": "device_key", (optional)"isAndroid":false (default true)}"
+     * Return: {"token": x}" where x if a token to save
+     * @note: the listen will timeout after six hours (and send a push notification).
+     * so you need to refresh the operation each six hours.
+     * @param session
+     */
+    void subscribe(const std::shared_ptr<restbed::Session>& session);
+    /**
+     * Unsubscribe to push notifications for an iOS or Android device.
+     * Method: UNSUBSCRIBE "/{InfoHash: .*}"
+     * Body: {"key": "device_key", "token": x} where x if the token to cancel
+     * Return: nothing
+     * @param session
+     */
+    void unsubscribe(const std::shared_ptr<restbed::Session>& session);
+    /**
+     * Send a push notification via a gorush push gateway
+     * @param key of the device
+     * @param json, the content to send
+     */
+    void sendPushNotification(const std::string& key, const Json::Value& json, bool isAndroid) const;
+
+    void cancelPushListen(const std::string& pushToken, const InfoHash& key, proxy::ListenToken token);
+
+
+#endif //OPENDHT_PUSH_NOTIFICATIONS
+
+    using clock = std::chrono::steady_clock;
+    using time_point = clock::time_point;
+
     std::thread server_thread {};
     std::unique_ptr<restbed::Service> service_;
     std::shared_ptr<DhtRunner> dht_;
 
+    std::mutex schedulerLock_;
+    std::condition_variable schedulerCv_;
+    Scheduler scheduler_;
+    std::thread schedulerThread_;
+
+    Sp<Scheduler::Job> printStatsJob_;
+
     // Handle client quit for listen.
     // NOTE: can be simplified when we will supports restbed 5.0
     std::thread listenThread_;
@@ -159,8 +240,26 @@ private:
         InfoHash hash;
         std::future<size_t> token;
     };
-    mutable std::vector<SessionToHashToken> currentListeners_;
+    std::vector<SessionToHashToken> currentListeners_;
+    std::mutex lockListener_;
     std::atomic_bool stopListeners {false};
+
+    struct PermanentPut;
+    struct SearchPuts;
+    std::map<InfoHash, SearchPuts> puts_;
+
+    mutable std::atomic<size_t> requestNum_ {0};
+    mutable std::atomic<time_point> lastStatsReset_ {time_point::min()};
+
+    const std::string pushServer_;
+
+#if OPENDHT_PUSH_NOTIFICATIONS
+    struct Listener;
+    struct PushListener;
+    std::mutex lockPushListeners_;
+    std::map<std::string, PushListener> pushListeners_;
+    proxy::ListenToken tokenPushNotif_ {0};
+#endif //OPENDHT_PUSH_NOTIFICATIONS
 };
 
 }
